/*
 * file CreateRecordCommand.java
 *
 * Licensed Materials - Property of IBM
 * Restricted Materials of IBM - you are allowed to copy, modify and 
 * redistribute this file as part of any program that interfaces with 
 * IBM Rational CM API.
 *
 * com.ibm.rational.stp.client.samples.CreateRecordCommand
 *
 * (C) Copyright IBM Corporation 2004, 2008.  All Rights Reserved.
 * Note to U.S. Government Users Restricted Rights:  Use, duplication or 
 * disclosure restricted by GSA ADP  Schedule Contract with IBM Corp.
 */
package com.ibm.rational.stp.client.samples;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.ListIterator;
import java.util.Map;
import java.util.TreeMap;

import javax.wvcm.PropertyNameList;
import javax.wvcm.WvcmException;
import javax.wvcm.PropertyNameList.PropertyName;
import javax.wvcm.PropertyRequestItem.PropertyRequest;

import com.ibm.rational.wvcm.stp.StpLocation;
import com.ibm.rational.wvcm.stp.StpProperty;
import com.ibm.rational.wvcm.stp.StpResource;
import com.ibm.rational.wvcm.stp.StpLocation.Namespace;
import com.ibm.rational.wvcm.stp.StpProvider.Domain;
import com.ibm.rational.wvcm.stp.cq.CqFieldValue;
import com.ibm.rational.wvcm.stp.cq.CqProvider;
import com.ibm.rational.wvcm.stp.cq.CqRecord;
import com.ibm.rational.wvcm.stp.cq.CqRecord.FieldName;

/**
 * A sample CM API application that demonstrates the use of the API to create
 * or update a ClearQuest record in a command line context.
 * <p>
 * The first argument is either the type of the record to be created or the name
 * of a record to be modified. Standard CM API selector syntax must be used.
 * The target user database must be specified in the repo field, but the domain
 * and the namespace fields may be omitted as they are understood to be <b>cq</b>
 * and <b>record</b>, respectively.
 * 
 * The command-line options supported are
 * <dl>
 * <dt>-user <i>user-name</i>
 * <dd>User logon name. Defaults to "admin"</dd>
 * <dt>-pass <i>password</i>
 * <dd>Password for the given user logon name. Defaults to ""</dd>
 * <dt>-action <i>action name</i>
 * <dd>The action to be used to update the record. Defaults to the default
 * Modify-type action defined for the record type.</dd>
 * <dt>[-set] <i>field-name</i> <i>field-value</i>
 * <dd>Specify a new value for a field</dd>
 * <dt>-setlist <i>field-name</i> <i>field-value</i>+
 * <dd>Specify a list of new values for a field. The value list is terminated
 * by the end of the command line or the next argument beginning with a hyphen</dd>
 * <dt>-show
 * <dd>Causes the current value of all fields to be displayed after the
 * operation completes. (If a record is identified by the first parameter, no
 * fields are specified, and no action is given, no operation is actually
 * performed: just the fields of the specified record are displayed.)</dd>
 * </dl>
 */
public class CreateRecordCommand
{
    /** Generally useful field meta-properties. * */
    private static final PropertyNameList FIELD_METAPROPERTIES =
        new PropertyNameList(CqFieldValue.VALUE,
                             CqFieldValue.NAME,
                             CqFieldValue.TYPE,
                             CqFieldValue.REQUIREDNESS);

    /** Property request for retrieving the user-friendly name of a resource */
    public static final PropertyRequest RECORD_REQUEST =
        new PropertyRequest(StpResource.USER_FRIENDLY_LOCATION,
                            CqRecord.ALL_FIELD_VALUES.nest(StpProperty.VALUE
                                .nest(FIELD_METAPROPERTIES)),
                            CqRecord.STATE_NAME,
                            CqRecord.DEFAULT_ACTION,
                            CqRecord.LEGAL_ACTIONS,
                            CqRecord.RECORD_CLASS);

    /** The CqProvider instance used by the class */
    private static CqProvider g_provider;

    /**
     * The database in which the record will be created determined from the
     * command line inputs
     */
    private static String g_repo = "";

    /**
     * Creates an StpLocation given a partial specification of its selector
     * 
     * @param name A String containing at least the name field of the selector
     * @param namespace The namespace required by context
     * @return An StpLocation for the given location
     * @throws WvcmException if the specified location is malformed.
     */
    static StpLocation normalize(Namespace namespace,
                                 String name) throws WvcmException
    {
        StpLocation loc = g_provider.stpLocation(name);

        if (loc.getDomain() == Domain.NONE)
            loc = loc.recomposeWithDomain(Domain.CLEAR_QUEST);

        if (loc.getNamespace() == Namespace.NONE) {
            loc = loc.recomposeWithNamespace(namespace);
        }

        if (g_repo.equals(""))
            g_repo = loc.getRepo();
        else if (loc.getRepo().equals(""))
            loc = loc.recomposeWithRepo(g_repo);

        return loc;
    }

    /**
     * Creates a new record using field values specified on the command line
     * 
     * @param args An array of String objects, each containing a command option
     *            or command option argument as described in the class overview
     *            comments.
     * @throws Exception If things go wrong.
     */
    public static void main(String[] args) throws Exception
    {
        run(args, RECORD_REQUEST);
        System.exit(0);
    }

    /**
     * Creates a new record using field values specified on the command line
     * 
     * @param args An array of String objects, each containing a command option
     *            or command option argument as described in the class overview
     *            comments.
     * @param wanted Target properties to be included in the returned proxy.
     *            Must include USER_FRIENDLY_LOCATION
     * @return A proxy for the created/modified record if the operation was
     *         successful; otherwise <b>null</b>
     * @throws Exception If things go wrong.
     */
    public static CqRecord run(String[] args,
                               PropertyRequest wanted) throws Exception
    {
        // Establish a session to ClearQuest
        g_provider = Utilities.getStaticProvider();

        boolean doShow = false;
        StpLocation actionLoc = null;
        ListIterator<String> argl = Arrays.asList(args).listIterator(0);
        StpLocation target = normalize(Namespace.RECORD, argl.next());
        boolean doCreate = target.getNameSegmentCount() == 1;
        CqRecord record =
            g_provider.cqRecord(doCreate? (StpLocation) target.child("new")
                                        : target);

        // Read and process the remaining command line arguments
        while (argl.hasNext()) {
            String arg = argl.next();

            if (arg.equals("-url"))
                g_provider.setServerUrl(argl.next());
            else if (arg.equals("-user"))
                Utilities.g_user = argl.next();
            else if (arg.equals("-password"))
                Utilities.g_pass = argl.next();
            else if (arg.equals("-action"))
                actionLoc = normalize(Namespace.ACTION, argl.next());
            else if (arg.equals("-show"))
                doShow = true;
            else if (arg.equals("-setlist")) {
                arg = argl.next();

                ArrayList<String> values = new ArrayList<String>();

                while (argl.hasNext()) {
                    String value = argl.next();

                    if (value.startsWith("-")) {
                        argl.previous();
                        break;
                    } else
                        values.add(value);
                }

                record.setProperty(new FieldName<Object>(arg), values);
            } else {
                if (arg.equals("-set"))
                    arg = argl.next();

                record.setProperty(new FieldName<String>(arg), argl.next());
            }
        }

        // Print the fields to be set
        PropertyName<?>[] updated =
            record.updatedPropertyNameList().getPropertyNames();

        for (PropertyName<?> name : updated)
            System.out.println(name.getName() + " => "
                + record.getProperty(name));

        // Read, create, or update record in database
        try {
            PropertyRequest rl = wanted;

            if (doCreate) {
                record = record.doCreateRecord(rl);
                System.out.println("Created "
                    + record.getUserFriendlyLocation());
            } else if (actionLoc != null) {
                // If record type was omitted from action location, insert it
                // now.
                // e.g.: input 'assign' became 'cq.action:assign@7.0.1/SAMPL';
                // now make it 'cq.action:Defect/assign@7.0.1/SAMPL'
                if (actionLoc.getNameSegmentCount() == 1)
                    actionLoc =
                        (StpLocation) actionLoc.parent().child(target.parent()
                            .lastSegment()).child(actionLoc.lastSegment());

                record =
                    (CqRecord) record.setAction(g_provider.cqAction(actionLoc))
                        .doWriteProperties(rl);
                System.out.println(actionLoc.lastSegment() + "(ed)" + target);
            } else {
                // If no action was specified, use doReadProperties. Modified
                // fields will be written if needed.
                record = (CqRecord) record.doReadProperties(wanted);
                System.out.println((updated.length > 0? "Updated" : "Found")
                    + record.getUserFriendlyLocation());
            }
        } catch (WvcmException ex) {
            System.out.println("Operation could not be completed: "
                + ex.getLocalizedMessage());

            // This will identify specific fields that were in error
            for (Throwable nested : ex.getNestedExceptions())
                System.out.println(nested.toString());

            return null;
        }

        if (doShow) {
            // Collect, sort, and print the field values
            Map<String, Object> report = new TreeMap<String, Object>();
            String pad = "";

            for (CqFieldValue field : record.getAllFieldValues()) {
                report.put(field.getName(), field.getValue());

                while (field.getName().length() > pad.length())
                    pad += " ";
            }

            for (String name : report.keySet()) {
                System.out.println(pad.substring(name.length()) + name + " : "
                    + showValue(pad + "   ", report.get(name)));
            }
        }

        return record;
    }

    private static Object showValue(String pad,
                                    Object value)
    {
        if (value != null) {
            StringBuffer sb = new StringBuffer();
            String prefix = "{";

            if (value instanceof Iterable) {
                for (Object item : (Iterable) value) {
                    sb.append(prefix + item);
                    prefix = "\n" + pad + " ";
                }

                sb.append("}");

                return sb.toString();
            } else if (value.getClass().isArray()) {
                prefix = "[";

                for (Object item : (Object[]) value) {
                    sb.append(prefix + item);
                    prefix = "\n" + pad + " ";
                }

                sb.append("]");

                return sb.toString();
            }
        }

        return value;
    }
}
