/*
 * file ProxyElement.java
 * 
 * Licensed Materials - Property of IBM
 * Restricted Materials of IBM - you are allowed to copy, modify and 
 * redistribute this file as part of any program that interfaces with 
 * IBM Rational CM API.
 *
 * com.ibm.rational.teamapi.scout.ProxyElement
 *
 * (C) Copyright IBM Corporation 2004, 2008.  All Rights Reserved.
 * Note to U.S. Government Users Restricted Rights:  Use, duplication or 
 * disclosure restricted by GSA ADP  Schedule Contract with IBM Corp.
 */
package com.ibm.rational.teamapi.scout;

import java.io.File;
import java.io.FileReader;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.wvcm.Folder;
import javax.wvcm.PropertyRequestItem;
import javax.wvcm.ProviderFactory;
import javax.wvcm.Resource;
import javax.wvcm.ResourceList;
import javax.wvcm.Workspace;
import javax.wvcm.WvcmException;
import javax.wvcm.PropertyNameList.PropertyName;
import javax.wvcm.PropertyRequestItem.PropertyRequest;
import javax.wvcm.ProviderFactory.Callback;
import javax.wvcm.ProviderFactory.Callback.Authentication;

import org.eclipse.jface.dialogs.InputDialog;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;

import com.ibm.rational.wvcm.stp.StpException;
import com.ibm.rational.wvcm.stp.StpLocation;
import com.ibm.rational.wvcm.stp.StpProvider;
import com.ibm.rational.wvcm.stp.StpResource;
import com.ibm.rational.wvcm.stp.StpException.StpReasonCode;
import com.ibm.rational.wvcm.stp.StpLocation.Namespace;
import com.ibm.rational.wvcm.stp.cc.CcDirectory;
import com.ibm.rational.wvcm.stp.cq.CqAttachmentFolder;
import com.ibm.rational.wvcm.stp.cq.CqDbSet;
import com.ibm.rational.wvcm.stp.cq.CqQueryFolder;
import com.ibm.rational.wvcm.stp.cq.CqRecord;
import com.ibm.rational.wvcm.stp.cq.CqRecordType;
import com.ibm.rational.wvcm.stp.cq.CqUserDb;


/**
 * A tree viewer model element for a CM API Resource. This model element also
 * implements the IPropertySource interface so that the element selected in the
 * tree viewer can be examined in the standard Eclipse property view.
 */
public class ProxyElement
    extends ResourceSource
{
    /**
     * Constructs a model element for a given resource. Used to implement the
     * public addChild methods.
     *
     * @param  parent  The parent of this element in the tree viewer. Is null
     *                 only for the (unseen) root of the tree.
     * @param  resource  The Resource proxy linking this element to the resource
     *                   being viewed. Must not be null except in the root
     *                   element of the tree.
     */
    private ProxyElement(ProxyElement parent,
                         StpResource  resource)
    {
        super(resource);
        m_parent = parent;
        m_namespace = resource.stpLocation().getNamespace();
    }

    /**
     * Constructs the root ProxyElement of a tree view.
     *
     * @param  shell  The display shell for this ProxyElement; used for context
     *                when requesting credentials from the user.
     *
     * @throws  Exception  if a CM API provider cannot be instantiated.
     */
    ProxyElement(Shell shell)
          throws Exception
    {
        super(null);

        m_parent = createProvider(shell);
        m_resource = null;
        m_children = new ArrayList<ProxyElement>();
    }

    /**
     * Adds a child resource to this element of the tree view model
     *
     * @param  child  The Resource proxy for the new child element.
     */
    void addChild(StpResource child)
    {
        m_children.add(new ProxyElement(this, child));
    }

    /**
     * Adds a child resource to this element of the tree view model
     *
     * @param  selectorString  A String containing the object selector for the
     *                         resource to be added as a child of this element.
     *
     * @throws  WvcmException  If the resource proxy can't be constructed
     */
    void addChild(String selectorString)
           throws WvcmException
    {
        StpProvider provider = getProvider();

        addChild((StpResource)provider.resource(provider.location(selectorString)));
    }

    /**
     * Removes a child model element from this model element.
     *
     * @param  child  The subordinate model element to be removed.
     *
     * @return  true if the child was an element and was removed; false
     *          otherwise.
     */
    boolean removeChild(Object child)
    {
        return m_children == null ? false : m_children.remove(child);
    }

    /**
     * Computes a String to identify this model element in the tree view.
     *
     * @return  A String containing the type and name of the resource.
     */
    public String getText()
    {
        if (isRoot())
            return "Namespaces";

        if (getParent().isRoot())
            return m_resource.location().string();

        try {
            return resourceType() + " " + m_resource.location().lastSegment();
        } catch (Throwable ex) {
            return m_resource.location().lastSegment() + "["
                + ex.getLocalizedMessage() + "]";
        }
    }

    /**
     * Returns the selector for this resource.
     *
     * @return  A String containing the image of this resource's selector.
     */
    public String getSelector()
    {
        if (m_resource == null)
            return "";

        return m_resource.toString();
    }

    /**
     * (non-Javadoc)
     *
     * @see  java.lang.Object#toString()
     *
     * @return  The Text for this ProxyElement.
     */
    public String toString()
    {
        return getText();
    }

    /**
     * For a resource in a given namespace, returns the PropertyName that 
     * identifies the property that serves as the bound member list of that 
     * resource in the namespace.
     * 
     * @param res
     * @param namespace
     * @return the property name
     * @throws WvcmException
     */
    PropertyName
    memberListProperty(StpResource res, 
                       Namespace namespace) 
    throws WvcmException
    {
        if (g_memberListPropertyMap.isEmpty()) {
            StpProvider provider = getProvider();
            
            for (int i = 0; i < table.length-1; ++i) {
                if (table[i+1] instanceof Class) {
                    HashMap<Object, Object> map = new HashMap<Object, Object>();
                    
                    g_memberListPropertyMap.put(table[i], map);
                    
                    for (; table[i+1] instanceof Class; i += 2)
                        map.put(provider.proxyType((Class<? extends Resource>)table[i+1]), 
                                (PropertyName)table[i+2]);
                }
            }
        }
        
        Map map = (Map)g_memberListPropertyMap.get(namespace);
        
        return (PropertyName)(map == null? null: map.get(res.proxyType()));
    }
    
    private static final HashMap<Object, Object> g_memberListPropertyMap = 
        new HashMap<Object, Object>();
    
    /**
     * For each namespace, maps the proxy classes in that namespace to the
     * property name for the property that defines the member list of that
     * resource type in the namespace.
     */
    private static final Object[] table = {
      Namespace.ACTION, 
          CqDbSet.class, CqDbSet.USER_DATABASES, 
          CqUserDb.class, CqUserDb.RECORD_TYPE_SET, 
          CqRecordType.class, CqRecordType.ACTION_LIST,
      Namespace.ACTIVITY,
      Namespace.ATTYPE,
      Namespace.BASELINE,
      Namespace.BRTYPE,
      Namespace.COMPONENT,
      Namespace.DB_SET,
      Namespace.DBID,
      Namespace.DYNAMIC_CHOICE_LIST,
          CqDbSet.class, CqDbSet.USER_DATABASES, 
          CqUserDb.class, CqUserDb.DYNAMIC_CHOICE_LISTS, 
      Namespace.ELTYPE,
      Namespace.FIELD_DEFINITION,
          CqDbSet.class, CqDbSet.USER_DATABASES, 
          CqUserDb.class, CqUserDb.RECORD_TYPE_SET, 
          CqRecordType.class, CqRecordType.FIELD_DEFINITIONS,
      Namespace.FILE,
      Namespace.FOLDER,
      Namespace.FORM,
          CqDbSet.class, CqDbSet.USER_DATABASES, 
          CqUserDb.class, CqUserDb.RECORD_TYPE_SET, 
      Namespace.GROUP,
      Namespace.HLINK,
      Namespace.HLTYPE,
      Namespace.HOOK,
          CqDbSet.class, CqDbSet.USER_DATABASES, 
          CqUserDb.class, CqUserDb.RECORD_TYPE_SET, 
          CqRecordType.class, CqRecordType.NAMED_HOOK_LIST,
      Namespace.HTTP,
      Namespace.HTTPS,
      Namespace.LBTYPE,
      Namespace.OID,
      Namespace.PNAME,
      Namespace.PNAME_IMPLIED,
      Namespace.POOL,
      Namespace.PROJECT,
      Namespace.PROJECT_CONFIGURATION,
      Namespace.QUERY,
          CqDbSet.class, CqDbSet.USER_DATABASES, 
          CqUserDb.class, CqUserDb.QUERY_FOLDER_ITEMS, 
          CqQueryFolder.class, CqQueryFolder.QUERY_FOLDER_ITEMS,
      Namespace.RECORD,
          CqDbSet.class, CqDbSet.USER_DATABASES, 
          CqUserDb.class, CqUserDb.RECORD_TYPE_SET, 
          CqRecord.class, CqRecord.ATTACHMENT_FOLDERS,
          CqAttachmentFolder.class, CqAttachmentFolder.ATTACHMENT_LIST,
      Namespace.REPLICA,
      Namespace.REPLICA_UUID,
      Namespace.REPO,
      Namespace.RPTYPE,
      Namespace.STREAM,
      Namespace.TRTYPE,
      Namespace.USER,
      Namespace.USER_DB,
      Namespace.VIEW_UUID,
      Namespace.VOB,
      Namespace.WORKSPACE,
         Workspace.class, Workspace.CHILD_LIST,
         CcDirectory.class, CcDirectory.CHILD_LIST,
      Namespace.NONE    // End of list
    };
    
    /**
     * Determines if this model element could be a folder.
     *
     * @return  true if the resource is a Folder proxy and it defines the
     *          CHILD_BINDING_LIST property; false otherwise.
     */
    boolean couldBeFolder()
    {
        if (m_resource != null && m_resource instanceof Folder) {
            Object obj = m_resource.lookupProperty(Folder.CHILD_LIST);

            if (obj == null || obj instanceof List)
                return true;

            if (obj instanceof StpException
                    && ((StpException)obj).getStpReasonCode() 
                        == StpReasonCode.PROPERTY_NOT_REQUESTED)
                return true;
        }

        return false;
    }

    /**
     * Determines if this model element should be considered a folder when
     * sorting and filtering the tree view.
     *
     * @return  true if the resource proxy is a folder proxy or if there is no
     *          resource associated with this element (i.e. the root of the
     *          tree).
     */
    boolean isFolder()
    {
        return m_resource == null || m_resource instanceof Folder;
    }

    /**
     * Determines if this model element is an empty folder;
     *
     * @return  true if the model element represents a folder with no children.
     */
    boolean isEmptyFolder()
    {
        return m_resource != null && m_resource instanceof Folder
                && getChildren().length == 0;
    }

    /**
     * Returns the children of the resource represented by this ProxyElement;
     * reading them from the CHILD_BINDING_LIST. The results are cached in this
     * element. A special case is made for selectors having no repository name.
     * This form of selector is interpreted as a request for one of the folder
     * lists available from the provider. Which method is used is based on the
     * namespace and repository type provided
     *
     * <table cellpadding=5 border=1>
     *   <tr>
     *     <th>Repository Type</th>
     *     <th>Namespace</th>
     *     <th>Method</th>
     *   </tr>
     *   <tr>
     *     <td>CLEAR_CASE</td>
     *     <td>REPOSITORY</td>
     *     <td>serverWorkspaceFolderList</td>
     *   </tr>
     *   <tr>
     *     <td>CLEAR_CASE</td>
     *     <td>VIEW<br>
     *       WORKSPACE</td>
     *     <td>clientWorkspaceFolderList</td>
     *   </tr>
     * </table>
     *
     * @return  An Array of model elements for the children.
     */
    Object[] getChildren()
    {
        if (m_children == null) {
            m_children = EMPTY_ARRAY;

            try {
                Object      obj = null;
                StpLocation selector = m_resource.stpLocation();

                if (selector.getRepo().length() == 0) {
                    StpProvider provider = getProvider();

                    if (selector.getDomain() == StpProvider.Domain.CLEAR_CASE) {
//                        if (selector.getNamespace() == Namespace.PROJECT) {
//                            obj =
//                                provider.ccProvider().(FOLDER_PROPERTIES);
//                        } else 
                            if (selector.getNamespace() == Namespace.WORKSPACE) {
                            obj =
                                provider.ccProvider().getClientViewList(FOLDER_PROPERTIES);
                        }
                    }
                }

                if (obj == null && couldBeFolder()) {
                    PropertyName memberList = memberListProperty(m_resource, 
                                                                 m_namespace);
                    obj = m_resource.lookupProperty(memberList);

                    if (obj instanceof StpException
                            && ((StpException)obj).getStpReasonCode()
                                    == StpReasonCode.PROPERTY_NOT_REQUESTED) {
                        final PropertyRequest FOLDER_PROPERTIES =
                            new PropertyRequest(new PropertyRequestItem[] {
                             StpResource.DISPLAY_NAME,
                             memberList.nest(new PropertyRequestItem[] {StpResource
                                                                .DISPLAY_NAME})
                         });
                        m_resource =
                            (StpResource)m_resource.doReadProperties(FOLDER_PROPERTIES);
                        obj = m_resource.lookupProperty(memberList);
                    }
                }

                if (obj != null) {
                    if (obj instanceof ResourceList) {
                        m_children = new ArrayList<ProxyElement>();

                        for (Iterator child = ((List)obj).iterator();
                                child.hasNext();)
                            addChild((StpResource)child.next());
                    } else if (obj instanceof List) {
                        m_children = new ArrayList<ProxyElement>();

                        for (Iterator child = ((List)obj).iterator();
                                child.hasNext();)
                            addChild((StpResource)child.next());
                    }
                }
            } catch (Throwable ex) {
                ex.printStackTrace();
            }
        }

        return m_children.toArray();
    }

    /**
     * Empties the cache of child elements; forcing them to be reread the next
     * time they are to be displayed.
     */
    void refresh()
    {
        m_children = null;
        m_resource.forgetProperty(Folder.CHILD_LIST);
    }

    /**
     * Returns whether or not this object is the root of the tree view.
     *
     * @return  true if this ProxyElement has not parent.
     */
    boolean isRoot()
    {
        return m_resource == null;
    }

    /**
     * The ProxyElement of which this is a child.
     *
     * @return  Returns the parent.
     */
    ProxyElement getParent()
    {
        return isRoot() ? null : (ProxyElement)m_parent;
    }

    /**
     * The Provider for this element of the tree.
     *
     * @return  A Provider object.
     */
    StpProvider getProvider()
    {
        return isRoot() ? (StpProvider)m_parent
                        : (StpProvider)m_resource.provider();
    }

    /**
     * At the root, the CM API provider for the tree; otherwise the parent
     * ProxyElement of this one
     */
    Object m_parent = null;

    /**
     * Constructs an instance of the CM API provider with an authenticator.
     *
     * @param  shell  The Shell to be used for display context when requesting
     *                credentials.
     *
     * @return  The instantiated Provider object
     *
     * @throws  Exception  If the Provider could not be instantiated
     */
    static StpProvider createProvider(final Shell shell)
                                   throws Exception
    {
        try {
            Callback callback =
                new ProviderFactory.Callback() {
                    private UnPw m_unpw;

                    Shell m_shell = shell;

                    public Authentication getAuthentication(final String  realm,
                                                            final int retryCount)
                    {
                        // Try to reuse last credentials on each new repository
                        if (m_unpw != null && retryCount == 0)
                            return m_unpw;

                        m_unpw = null;

                        // Since we may be called from a non-UI thread, we need
                        // to take measures to ensure the dialog comes up on the
                        // UI thread.
                        Display  display = m_shell.getDisplay();
                        Runnable runnable =
                            new Runnable() {
                                public void run()
                                {
                                    InputDialog dialog =
                                        new InputDialog(m_shell,
                                                        "CM API Scout Login",
                                                        "Enter Username '+' Password for "
                                                        + realm + " ["
                                                        + retryCount + "]",
                                                        "admin+", null);

                                    try {
                                        if (dialog.open() == InputDialog.OK) {
                                            String unpw = dialog.getValue();

                                            if (unpw.startsWith("@")) {
                                                File file =
                                                    new File(unpw.substring(1));

                                                FileReader reader =
                                                    new FileReader(file);
                                                char[]     buf = new char[100];
                                                int        count =
                                                    reader.read(buf);

                                                unpw =
                                                    new String(buf, 0, count);
                                                reader.close();
                                            }

                                            m_unpw =
                                                new UnPw(unpw.split("\\+", -2));
                                        }
                                    } catch (Throwable t) {
                                        t.printStackTrace();
                                    }
                                }
                            };

                        display.syncExec(runnable);

                        if (m_unpw == null)
                            throw new UnsupportedOperationException("No credentials available");

                        return m_unpw;
                    }
                };

            // Instantiate a Provider
            return (StpProvider)ProviderFactory
                .createProvider(StpProvider.PROVIDER_CLASS, callback);
        } catch (InvocationTargetException ite) {
            System.out.println("*** " + ite.getLocalizedMessage());

            StpException ex = (StpException)ite.getTargetException();
            Throwable[]  nested = ex.getNestedExceptions();

            for (int i = 0; i < nested.length; ++i)
                System.out.println("***  " + nested[i].getLocalizedMessage());

            throw ex;
        }
    }

    /**
     * A simple Authentication object in which the username and password
     * obtained from the user is cached for use by the CM API.
     */
    static class UnPw
        implements Authentication
    {
        /**
         * Creates a new UnPw object.
         *
         * @param  unpw  The user name and password stored in a 1 or 2 element
         *               array
         */
        UnPw(String[] unpw)
        {
            m_data = unpw;
        }

        /**
         * loginName
         *
         * @return  the login name for this user
         */
        public String loginName()
        {
            return m_data[0];
        }

        /**
         * password
         *
         * @return  The password for this user
         */
        public String password()
        {
            return m_data.length > 1 ? m_data[1] : "";
        }
        ;

        /** The user-name and password array */
        private String[] m_data;
    }

    /**
     * Returns a String suitable for displaying the type of resource as
     * determined by its proxy class.
     *
     * @return  A String containing the simple name of the most derived CM API
     *          interface implemented by the proxy of this ProxyElement.
     */
    String resourceType()
    {
        if (isRoot() || getParent().isRoot())
            return "";

        Class  proxyClass = m_resource.getClass();
        String name = (String)g_typeMap.get(proxyClass);

        if (name == null) {
            Class<?>[] interfaces = proxyClass.getInterfaces();
            Class<?>   choice = StpResource.class;

            for (int i = 0; i < interfaces.length; ++i) {
                name = interfaces[i].getName();

                if (name.startsWith("com.ibm.rational.wvcm.")
                        || name.startsWith("javax.wvcm")) {
                    // Within the CM API, select the most derived interface
                    // this resource implements
                    if (choice.isAssignableFrom(interfaces[i]))
                        choice = interfaces[i];
                }
            }

            name = choice.getName();
            name = name.substring(1 + name.lastIndexOf('.'));

            g_typeMap.put(proxyClass, name);
        }

        return name;
    }

    /** The children (virtual CHILD_BINDING_LIST) of this ProxyElement. */
    List<ProxyElement> m_children = null;

    /** The namespace that is being traversed through this element */
    Namespace m_namespace;
    
    /** A constant used for empty results */
    private static List<ProxyElement> EMPTY_ARRAY = new ArrayList<ProxyElement>();

    /** The properties requested from a Folder */
    private static final PropertyRequest FOLDER_PROPERTIES =
        new PropertyRequest(new PropertyRequestItem[] {
                                 StpResource.DISPLAY_NAME,
                                 Folder.CHILD_LIST.nest(new PropertyRequestItem[] {
                                                 StpResource.DISPLAY_NAME})
                             });

    /** Maps proxy classes to a string used to identify their type */
    private static final Map<Class<?>, String> g_typeMap = 
        new HashMap<Class<?>, String>();
}
